from typing import TYPE_CHECKING, Optional

import qtawesome as qta
from PySide6.QtCore import QAbstractItemModel, QModelIndex, Qt
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QComboBox, QLabel, QMenu

from angrmanagement.config import Conf
from angrmanagement.logic.debugger import Debugger, DebuggerListManager, DebuggerManager, DebuggerWatcher

from .toolbar import Toolbar, ToolbarAction, ToolbarSplitter

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.main_window import MainWindow
    from angrmanagement.ui.workspace import Workspace


class AvailableDebuggersModel(QAbstractItemModel):
    """
    Data provider for available debuggers combo box.
    """

    def __init__(self, workspace: "Workspace"):
        super().__init__()
        self.debugger_mgr: DebuggerManager = workspace.main_instance.debugger_mgr
        self.debugger_list_mgr: DebuggerListManager = workspace.main_instance.debugger_list_mgr
        self.last_str = {}

    def rowCount(self, parent):  # pylint:disable=unused-argument
        return len(self.debugger_list_mgr.debugger_list) + 1

    def columnCount(self, parent):  # pylint:disable=unused-argument,no-self-use
        return 1

    def index(self, row, col, parent):  # pylint:disable=unused-argument
        return self.createIndex(row, col, None)

    def parent(self, index):  # pylint:disable=unused-argument,no-self-use
        return QModelIndex()

    def data(self, index, role):
        if not index.isValid() or role != Qt.DisplayRole:
            return None
        row = index.row()
        dbg = self.index_to_debugger(row)
        self.last_str[row] = "No Debugger" if dbg is None else str(dbg)
        return self.last_str[row]

    def debugger_to_index(self, dbg: Optional[Debugger]) -> int:
        return 0 if dbg is None else (self.debugger_list_mgr.debugger_list.index(dbg) + 1)

    def index_to_debugger(self, index: int) -> Optional[Debugger]:
        return None if index == 0 else (self.debugger_list_mgr.debugger_list[index - 1])


class DebugToolbar(Toolbar):
    """
    Debugger Control Toolbar
    TODO: decouple this from MainWindow and workspace.instance (main_instance)
    """

    def __init__(self, main_window: "MainWindow"):
        super().__init__(main_window, "DebugToolbar")
        self.workspace: Workspace = main_window.workspace
        self.instance: Instance = self.workspace.main_instance

        self._cont_backward_act = ToolbarAction(
            qta.icon("fa5s.fast-backward", color=Conf.palette_buttontext),
            "Continue-Backward",
            "Reverse-Continue",
            self._on_cont_backward,
        )
        self._step_backward_act = ToolbarAction(
            qta.icon("fa5s.step-backward", color=Conf.palette_buttontext),
            "Step-Backward",
            "Reverse-Step",
            self._on_step_backward,
        )
        self._cont_act = ToolbarAction(
            qta.icon("fa5s.play", color=Conf.palette_buttontext), "Continue", "Continue", self._on_cont
        )
        self._halt_act = ToolbarAction(
            qta.icon("fa5s.pause", color=Conf.palette_buttontext), "Halt", "Halt", self._on_halt
        )
        self._step_act = ToolbarAction(
            qta.icon("fa5s.step-forward", color=Conf.palette_buttontext), "Step", "Step", self._on_step
        )
        self._step_over_act = ToolbarAction(
            qta.icon("fa5s.share", color=Conf.palette_buttontext), "Step Over", "Step Over", self._on_step_over
        )

        self._start_act = ToolbarAction(
            qta.icon("fa5s.running", color=Conf.palette_buttontext), "Launch", "New Debugger", self._on_start
        )
        self._stop_act = ToolbarAction(
            qta.icon("fa5s.stop-circle", color=Conf.palette_buttontext), "Stop", "Stop Debugging", self._on_stop
        )

        self.actions = [
            self._cont_backward_act,
            self._step_backward_act,
            self._cont_act,
            self._halt_act,
            self._step_act,
            self._step_over_act,
            ToolbarSplitter(),
            self._start_act,
            self._stop_act,
        ]
        self.qtoolbar()

        self._cached.visibilityChanged.connect(self._on_visibility_changed)

        self._dbg_list_mgr = self.instance.debugger_list_mgr
        self._dbg_mgr = self.instance.debugger_mgr

        self._new_dbg_menu = QMenu(self._cached)

        self._new_dbg_sim = QAction("New simulation...", self._new_dbg_menu)
        self._new_dbg_sim.triggered.connect(self.workspace.main_window.open_newstate_dialog)
        self._new_dbg_menu.addAction(self._new_dbg_sim)

        self._new_dbg_trace = QAction("New trace debugger", self._new_dbg_menu)
        self._new_dbg_trace.triggered.connect(self.workspace.create_trace_debugger)
        self._new_dbg_menu.addAction(self._new_dbg_trace)

        self._new_dbg_menu.aboutToShow.connect(self._update_new_dbg_list)
        self._cached_actions[self._start_act].setMenu(self._new_dbg_menu)
        self._update_new_dbg_list()

        self._dbg_model = AvailableDebuggersModel(self.workspace)
        self._dbg_combo = QComboBox()
        self._dbg_combo.setMinimumWidth(250)
        self._dbg_combo.setModel(self._dbg_model)
        self._dbg_combo.activated.connect(self._select_dbg_by_index)
        self._cached.addWidget(self._dbg_combo)
        self._update_dbg_list_combo()

        self._run_lbl = QLabel()
        self._run_lbl.setText("")
        self._cached.addWidget(self._run_lbl)

        self._dbg_watcher: Optional[DebuggerWatcher] = None

    def _on_visibility_changed(self, visible: bool):
        if visible:
            self.instance.debugger_list_mgr.debugger_list.am_subscribe(self._update_dbg_list_combo)
            self._dbg_watcher = DebuggerWatcher(self._dbg_state_changed, self._dbg_mgr.debugger)
            self._dbg_state_changed()
        else:
            self._dbg_watcher.shutdown()
            self._dbg_watcher = None

    def _select_dbg_by_index(self, index: int):
        dbg = self._dbg_model.index_to_debugger(index)
        self._dbg_mgr.set_debugger(dbg)

    def _dbg_state_changed(self):
        self._update_dbg_list_combo()
        self._update_state()

    def _select_current_dbg_in_combo(self, *args, **kwargs):  # pylint:disable=unused-argument
        dbg = self._dbg_mgr.debugger.am_obj
        self._dbg_combo.setCurrentIndex(self._dbg_model.debugger_to_index(dbg))

    def _update_new_dbg_list(self):
        self._new_dbg_sim.setDisabled(self.instance.project.am_none)
        self._new_dbg_trace.setDisabled(self.instance.current_trace.am_none)

    def _update_dbg_list_combo(self, *args, **kwargs):  # pylint:disable=unused-argument
        dl = self.instance.debugger_list_mgr.debugger_list
        self._dbg_combo.setEnabled(len(dl) > 0)
        self._select_current_dbg_in_combo()
        self._dbg_model.layoutChanged.emit()
        self._dbg_combo.update()

    def _update_state(self):
        dbg = self._dbg_mgr.debugger.am_obj
        dbg_active = dbg is not None

        def q(a):
            return self._cached_actions[a]

        q(self._step_act).setEnabled(dbg_active and dbg.can_step_forward)
        q(self._step_over_act).setEnabled(dbg_active and dbg.can_step_forward)
        q(self._step_backward_act).setEnabled(dbg_active and dbg.can_step_backward)
        q(self._cont_backward_act).setEnabled(dbg_active and dbg.can_continue_backward)
        q(self._cont_act).setEnabled(dbg_active and dbg.can_continue_forward)
        q(self._cont_act).setVisible(not (dbg_active and dbg.can_halt))
        q(self._halt_act).setEnabled(dbg_active and dbg.can_halt)
        q(self._halt_act).setVisible(dbg_active and dbg.can_halt)
        q(self._start_act).setEnabled(True)
        q(self._stop_act).setEnabled(dbg_active and dbg.can_stop)

        if dbg_active:
            self._run_lbl.setText(dbg.state_description)
        else:
            self._run_lbl.setText("")

    def _on_start(self):
        self._cached.widgetForAction(self._cached_actions[self._start_act]).showMenu()

    def _on_stop(self):
        self._dbg_mgr.debugger.stop()
        self._dbg_list_mgr.remove_debugger(self._dbg_mgr.debugger.am_obj)

    def _on_cont(self):
        self._dbg_mgr.debugger.continue_forward()

    def _on_cont_backward(self):
        self._dbg_mgr.debugger.continue_backward()

    def _on_halt(self):
        self._dbg_mgr.debugger.halt()

    def _on_step(self):
        self._dbg_mgr.debugger.step_forward()

    def _on_step_over(self):
        b = self._dbg_mgr.debugger.simstate.block()
        until_addr = b.instruction_addrs[0] + b.size if b.instructions == 1 and b.vex.jumpkind == "Ijk_Call" else None
        self._dbg_mgr.debugger.step_forward(until_addr=until_addr)

    def _on_step_backward(self):
        self._dbg_mgr.debugger.step_backward()



from .job import Job


class SimgrStepJob(Job):
    def __init__(self, simgr, callback=None, until_branch=False, step_callback=None):
        super().__init__("Simulation manager stepping")
        self._simgr = simgr
        self._callback = callback
        self._until_branch = until_branch
        self._step_callback = step_callback

    def _run(self, inst):
        if self._until_branch:
            orig_len = len(self._simgr.active)
            if orig_len > 0:
                while len(self._simgr.active) == orig_len:
                    self._simgr.step(step_func=self._step_callback)
                    self._simgr.prune()
        else:
            self._simgr.step(step_func=self._step_callback, num_inst=1)
            self._simgr.prune()

        return self._simgr

    def finish(self, inst, result):
        super().finish(inst, result)
        if self._callback is not None:
            self._callback(result)

    def __repr__(self):
        if self._until_branch:
            return "Stepping %r until branch" % self._simgr
        else:
            return "Stepping %r" % self._simgr

    @classmethod
    def create(cls, simgr, **kwargs):
        def callback(result):
            simgr.am_event(src="job_done", job="step", result=result)

        return cls(simgr, callback=callback, **kwargs)



from typing import TYPE_CHECKING, Any, Optional

from PySide6.QtCore import QAbstractTableModel, QSize, Qt
from PySide6.QtGui import QBrush, QFont
from PySide6.QtWidgets import QAbstractItemView, QHeaderView, QTableView, QVBoxLayout

from angrmanagement.config import Conf
from angrmanagement.logic.debugger import DebuggerWatcher

from .view import BaseView

if TYPE_CHECKING:
    import angr
    import PySide6
    from archinfo import Register


class QRegisterTableModel(QAbstractTableModel):
    """
    Register table model.
    """

    Headers = ["Name", "Value"]
    COL_REGISTER = 0
    COL_VALUE = 1

    def __init__(self, log_widget: "QRegisterTableWidget" = None):
        super().__init__()
        self._log_widget = log_widget
        self.state: angr.SimState = None
        self._last_state: angr.SimState = None

    def _filtered_register_list(self):
        return [reg for reg in self.state.arch.register_list if reg.general_purpose]

    def rowCount(self, parent: "PySide6.QtCore.QModelIndex" = ...) -> int:  # pylint:disable=unused-argument
        return 0 if self.state is None else len(self._filtered_register_list())

    def columnCount(self, parent: "PySide6.QtCore.QModelIndex" = ...) -> int:  # pylint:disable=unused-argument
        return len(self.Headers)

    def headerData(
        self, section: int, orientation: "PySide6.QtCore.Qt.Orientation", role: int = ...
    ) -> Any:  # pylint:disable=unused-argument
        if role != Qt.DisplayRole:
            return None
        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index: "PySide6.QtCore.QModelIndex", role: int = ...) -> Any:
        if not index.isValid():
            return None
        row = index.row()
        reg = self._filtered_register_list()[row]
        col = index.column()
        if role == Qt.DisplayRole:
            return self._get_column_text(reg, col)
        elif role == Qt.ForegroundRole:
            return QBrush(Qt.red) if self._did_data_change(reg) else None
        else:
            return None

    def _get_column_text(self, reg: "Register", col: int) -> Any:
        mapping = {
            QRegisterTableModel.COL_REGISTER: lambda x: x.name,
            QRegisterTableModel.COL_VALUE: lambda x: repr(self.state.regs.get(x.name)),
        }
        func = mapping.get(col)
        if func is None:
            return None
        return func(reg)

    def _did_data_change(self, reg: "Register") -> bool:
        if self._last_state is None:
            return False
        different = self.state.solver.eval(self.state.regs.get(reg.name) != self._last_state.regs.get(reg.name))
        return different

    def update_state(self, state):
        self._last_state = self.state.copy() if self.state else None
        self.state = None if state is None else state


class QRegisterTableWidget(QTableView):
    """
    Register table widget.
    """

    def __init__(self, register_view, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_view = register_view

        hheader = self.horizontalHeader()
        hheader.setVisible(True)
        hheader.setStretchLastSection(True)

        vheader = self.verticalHeader()
        vheader.setVisible(False)
        vheader.setDefaultSectionSize(20)

        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)

        self.model: QRegisterTableModel = QRegisterTableModel(self)
        self.setModel(self.model)

        font = QFont(Conf.disasm_font)
        self.setFont(font)

        hheader.setSectionResizeMode(0, QHeaderView.ResizeToContents)

        self._dbg_manager = register_view.instance.debugger_mgr
        self._dbg_watcher = DebuggerWatcher(self._on_debugger_state_updated, self._dbg_manager.debugger)
        self._on_debugger_state_updated()

    #
    # Events
    #

    def closeEvent(self, event):
        self._dbg_watcher.shutdown()
        super().closeEvent(event)

    def _on_debugger_state_updated(self):
        dbg = self._dbg_manager.debugger
        self.model.update_state(None if dbg.am_none else dbg.simstate)
        self.model.layoutChanged.emit()


class RegistersView(BaseView):
    """
    Register table view.
    """

    def __init__(self, workspace, instance, default_docking_position, *args, **kwargs):
        super().__init__("registers", workspace, instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Registers"
        self._tbl_widget: Optional[QRegisterTableWidget] = None
        self._init_widgets()
        self.reload()

        self.width_hint = 500
        self.height_hint = 0
        self.updateGeometry()

    def reload(self):
        pass

    @staticmethod
    def minimumSizeHint(*args, **kwargs):  # pylint:disable=unused-argument
        return QSize(200, 200)

    def _init_widgets(self):
        vlayout = QVBoxLayout()
        vlayout.setContentsMargins(0, 0, 0, 0)
        self._tbl_widget = QRegisterTableWidget(self)
        vlayout.addWidget(self._tbl_widget)
        self.setLayout(vlayout)



from collections import OrderedDict
from typing import TYPE_CHECKING, Optional

from angr.analyses.decompiler.structured_codegen.c import CConstruct, CFunction, CFunctionCall, CStructField, CVariable
from angr.sim_type import NamedTypeMixin, SimType, SimTypePointer, TypeRef
from PySide6.QtGui import Qt
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QPushButton,
    QVBoxLayout,
)

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function

    from angrmanagement.ui.views.code_view import CodeView


class NodeNameBox(QLineEdit):
    """
    QLineEdit that validates node names.
    """

    def __init__(self, textchanged_callback, parent=None):
        super().__init__(parent)

        self.textChanged.connect(textchanged_callback)

    @property
    def name(self):
        text = self.text()
        if self._is_valid_node_name(text):
            return text.strip()
        return None

    @staticmethod
    def _is_valid_node_name(name):
        stripped = name.strip()
        return stripped and " " not in stripped


class RenameNode(QDialog):
    """
    Dialog for renaming a node.
    """

    def __init__(
        self,
        code_view: Optional["CodeView"] = None,
        node: Optional[CConstruct] = None,
        parent=None,
        func: Optional["Function"] = None,
    ):
        super().__init__(parent)

        # initialization
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        self._code_view = code_view
        self._node = node
        self._func: Optional[Function] = func

        self._name_box: NodeNameBox = None
        self._status_label = None
        self._ok_button: QPushButton = None
        self._suggestion_box: QListWidget = None

        self.setWindowTitle("Rename")

        self.main_layout = QVBoxLayout()

        self._init_widgets()

        if not isinstance(self._node, CVariable):
            self._suggestion_box.setEnabled(False)
        else:
            if self._node.unified_variable is not None:
                if not self._node.unified_variable.candidate_names:
                    self._suggestion_box.setEnabled(False)
                else:
                    for candidate in self._node.unified_variable.candidate_names:
                        self._suggestion_box.addItem(candidate)

        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _init_widgets(self):
        # name label

        name_label = QLabel(self)
        name_label.setText("New name")

        name_box = NodeNameBox(self._on_name_changed, self)
        if self._node is not None:
            # parse node type, either a Function header or a Variable.
            if isinstance(self._node, CVariable) and self._node.unified_variable and self._node.unified_variable.name:
                name_box.setText(self._node.unified_variable.name)
            elif isinstance(self._node, CVariable) and not self._node.variable.region:
                name_box.setText(self._node.variable.name)
            elif isinstance(self._node, CFunction) and self._node.name:
                name_box.setText(self._node.name)
            elif isinstance(self._node, CFunctionCall):
                name_box.setText(self._node.callee_func.name)
            elif isinstance(self._node, CStructField):
                name_box.setText(self._node.field)

            name_box.selectAll()
        self._name_box = name_box

        label_layout = QHBoxLayout()
        label_layout.addWidget(name_label)
        label_layout.addWidget(name_box)
        self.main_layout.addLayout(label_layout)

        # suggestions
        suggest_label = QLabel(self)
        suggest_label.setText("Suggestions")

        suggestion_box = QListWidget()
        self._suggestion_box = suggestion_box
        suggestion_layout = QHBoxLayout()
        suggestion_layout.addWidget(suggest_label)
        suggestion_layout.addWidget(suggestion_box)
        self.main_layout.addLayout(suggestion_layout)

        # status label
        status_label = QLabel(self)
        self.main_layout.addWidget(status_label)
        self._status_label = status_label

        # buttons
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        self._ok_button = buttons.button(QDialogButtonBox.Ok)
        self._ok_button.setEnabled(False)
        self.main_layout.addWidget(buttons)

    #
    # Event handlers
    #

    def _on_name_changed(self, new_text):  # pylint:disable=unused-argument
        if self._name_box is None:
            # initialization is not done yet
            return

        if self._name_box.name is None:
            # the variable name is invalid
            self._status_label.setText("Invalid")
            self._status_label.setProperty("class", "status_invalid")
            self._ok_button.setEnabled(False)
        else:
            self._status_label.setText("Valid")
            self._status_label.setProperty("class", "status_valid")
            self._ok_button.setEnabled(True)

        self._status_label.style().unpolish(self._status_label)
        self._status_label.style().polish(self._status_label)

    def _on_ok_clicked(self):
        node_name = self._name_box.name
        if node_name is not None and self._code_view is not None and self._node is not None:
            # need workspace for altering callbacks of changes
            workspace = self._code_view.workspace
            code_kb = self._code_view.codegen.kb

            # stack variable
            if isinstance(self._node, CVariable) and self._node.unified_variable is not None:
                # sanity check that we are a stack var
                if hasattr(self._node.variable, "offset") and self._node.variable.offset is not None:
                    workspace.plugins.handle_stack_var_renamed(
                        self._func,
                        self._node.variable.offset,
                        self._node.variable.name,
                        node_name,
                    )

                self._node.unified_variable.name = node_name
                self._node.unified_variable.renamed = True

            # global variable
            elif isinstance(self._node, CVariable) and not self._node.variable.region:
                workspace.plugins.handle_global_var_renamed(
                    self._node.variable.addr, self._node.variable.name, node_name
                )

                self._code_view.instance.kb.labels[self._node.variable.addr] = node_name
                self._node.variable.name = node_name
                self._node.variable.renamed = True

            # function arg
            elif isinstance(self._node, CVariable):
                workspace.plugins.handle_func_arg_renamed(
                    code_kb.functions[self._node.codegen.cfunc.addr], 0, self._node.variable.name, node_name
                )

                self._node.variable.name = node_name
                self._node.variable.renamed = True

            # function name
            elif isinstance(self._node, CFunction):
                workspace.plugins.handle_function_renamed(
                    code_kb.functions[self._node.codegen.cfunc.addr], self._node.name, node_name
                )

                code_kb.functions.get_by_addr(self._node.addr).name = node_name
                self._node.name = node_name
                self._node.demangled_name = node_name

            # function renaming (as a call)
            elif isinstance(self._node, CFunctionCall):
                if self._node.callee_func is not None:
                    workspace.plugins.handle_function_renamed(
                        code_kb.functions[self._node.codegen.cfunc.addr], self._node.callee_func.name, node_name
                    )

                    self._node.callee_func.name = node_name

            # struct renaming
            elif isinstance(self._node, CStructField):
                # TODO add a better callback
                # TODO prevent name duplication. reuse logic from CTypeEditor?
                # TODO if this is a temporary struct, make it permanent and add it to kb.types
                fields = [
                    (node_name if n == self._node.field else n, t) for n, t in self._node.struct_type.fields.items()
                ]
                self._node.struct_type.fields = OrderedDict(fields)

                workspace.plugins.handle_struct_changed(self._node.field, node_name)

                self._node.field = node_name

            elif isinstance(self._node, SimType):
                ty = self._node
                ref = None
                while True:
                    if isinstance(ty, TypeRef):
                        ref = ty
                        ty = ty.type
                    elif isinstance(ty, SimTypePointer):
                        ref = None
                        ty = ty.pts_to
                    else:
                        break
                if isinstance(ty, NamedTypeMixin):
                    ty.name = node_name
                    if ref is not None:  # update the typeref as well
                        ref._name = node_name
                else:
                    ty.label = node_name

            self._code_view.codegen.am_event()
            self.close()



import logging
from typing import TYPE_CHECKING, Dict, List, Optional

import angr
import archinfo
import cle
from angr.angrdb import AngrDB
from PySide6.QtWidgets import QMessageBox

from angrmanagement.logic.threads import gui_thread_schedule
from angrmanagement.ui.dialogs import LoadBinary

from .job import Job

try:
    import archr
except ImportError:
    archr = None


if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase


_l = logging.getLogger(__name__)


class LoadTargetJob(Job):
    """
    Job to load archr target and angr project.
    """

    def __init__(self, target, on_finish=None):
        super().__init__("Loading target", on_finish=on_finish)
        self.target = target

    def _run(self, inst):
        self._progress_callback(5)
        with self.target.build().start() as t:
            self._progress_callback(10)
            dsb = archr.arsenal.DataScoutBow(t)
            apb = archr.arsenal.angrProjectBow(t, dsb)
            partial_ld = apb.fire(return_loader=True, perform_relocations=False, load_debug_info=False)
            self._progress_callback(50)
            load_options = gui_thread_schedule(LoadBinary.run, (partial_ld,))
            if load_options is None:
                return

            # Create the project, load it, then record the image name on success
            proj = apb.fire(use_sim_procedures=True, load_options=load_options)
            self._progress_callback(95)
            inst._reset_containers()
            inst.project = proj
            inst.project.am_event()


class LoadBinaryJob(Job):
    """
    Job to display binary load dialog and create angr project.
    """

    def __init__(self, fname, load_options=None, on_finish=None):
        super().__init__("Loading file", on_finish=on_finish)
        self.load_options = load_options or {}
        self.fname = fname

    def _run(self, inst):
        self._progress_callback(5)

        load_as_blob = False
        load_with_libraries = True

        partial_ld = None
        try:
            # Try automatic loading
            partial_ld = cle.Loader(
                self.fname, perform_relocations=False, load_debug_info=False, main_opts={"ignore_missing_arch": True}
            )
        except archinfo.arch.ArchNotFound:
            _l.warning("Could not identify binary architecture.")
            partial_ld = None
            load_as_blob = True
        except cle.CLECompatibilityError:
            # Continue loading as blob
            load_as_blob = True
        except cle.CLEError:
            # try loading as a single binary without libraries
            _l.warning("Failed to load the binary with libraries.")
            load_with_libraries = False

        if partial_ld is None and not load_with_libraries:
            try:
                # Try loading as blob; dummy architecture (x86) required, user will select proper arch
                partial_ld = cle.Loader(
                    self.fname,
                    perform_relocations=False,
                    load_debug_info=False,
                    auto_load_libs=False,
                )
            except (cle.CLECompatibilityError, cle.CLEError):
                # failed to load without libraries
                load_as_blob = True

        if partial_ld is None and load_as_blob:
            try:
                # Try loading as blob; dummy architecture (x86) required, user will select proper arch
                partial_ld = cle.Loader(self.fname, main_opts={"backend": "blob", "arch": "x86"})
            except cle.CLECompatibilityError:
                # Failed to load executable, even as blob!
                gui_thread_schedule(LoadBinary.binary_loading_failed, (self.fname,))
                return

        self._progress_callback(50)
        new_load_options, simos = gui_thread_schedule(
            LoadBinary.run, (partial_ld, partial_ld.main_object.__class__, partial_ld.main_object.os)
        )
        if new_load_options is None:
            return

        engine = None
        if hasattr(new_load_options["arch"], "pcode_arch"):
            engine = angr.engines.UberEnginePcode

        self.load_options.update(new_load_options)

        proj = angr.Project(self.fname, load_options=self.load_options, engine=engine, simos=simos)
        self._progress_callback(95)

        def callback():
            inst._reset_containers()
            inst.project.am_obj = proj
            inst.project.am_event()

        gui_thread_schedule(callback, ())


class LoadAngrDBJob(Job):
    """
    Load an angr database file and return a new angr project.
    """

    def __init__(
        self,
        file_path: str,
        kb_names: List[str],
        other_kbs: Optional[Dict[str, "KnowledgeBase"]] = None,
        extra_info: Optional[Dict] = None,
        on_finish=None,
    ):
        super().__init__("Loading angr database", on_finish=on_finish)
        self.file_path = file_path
        self.kb_names = kb_names
        self.other_kbs = other_kbs
        self.extra_info = extra_info
        self.blocking = True

        self.project = None

    def _run(self, inst):
        self._progress_callback(5)

        angrdb = AngrDB()
        try:
            proj = angrdb.load(
                self.file_path, kb_names=self.kb_names, other_kbs=self.other_kbs, extra_info=self.extra_info
            )
        except angr.errors.AngrIncompatibleDBError as ex:
            _l.critical("Failed to load the angr database because of compatibility issues.", exc_info=True)
            gui_thread_schedule(
                QMessageBox.critical,
                (None, "Error", f"Failed to load the angr database because of compatibility issues.\nDetails: {ex}"),
            )
            return
        except angr.errors.AngrDBError as ex:
            _l.critical("Failed to load the angr database because of compatibility issues.", exc_info=True)
            gui_thread_schedule(
                QMessageBox.critical, (None, "Error", f"Failed to load the angr database.\nDetails: {ex}")
            )
            _l.critical("Failed to load the angr database.", exc_info=True)
            return

        self.project = proj

        self._progress_callback(100)



from typing import TYPE_CHECKING, List

from angr.knowledge_plugins.variables.variable_access import VariableAccess
from angr.knowledge_plugins.xrefs.xref import XRef, XRefType
from PySide6.QtCore import QAbstractTableModel, Qt
from PySide6.QtWidgets import QAbstractItemView, QHeaderView, QTableView

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance


class XRefMode:
    Variable = "variable"
    Address = "address"


class QXRefModel(QAbstractTableModel):
    HEADER = []

    def __init__(self, addr, instance, view):
        super().__init__()

        self.addr = addr
        self.instance = instance
        self.view = view

    @property
    def xrefs(self) -> List[VariableAccess]:
        return self.view.items

    @xrefs.setter
    def xrefs(self, v):
        self.view.items = v

    def __len__(self):
        return len(self.xrefs)

    def rowCount(self, parent=None):
        return len(self.xrefs)

    def columnCount(self, parent=None):
        return len(self.HEADER)

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole:
            if section < len(self.HEADER):
                return self.HEADER[section][0]
        elif role == Qt.InitialSortOrderRole:
            return Qt.AscendingOrder

        return None

    def data(self, index, role):
        if not index.isValid():
            return None

        row = index.row()
        if row >= len(self.xrefs):
            return None

        col = index.column()
        xref = self.xrefs[row]

        if role == Qt.DisplayRole:
            return self._get_column_text(xref, col)

        elif role == Qt.FontRole:
            return Conf.tabular_view_font

        return None

    def sort(self, column, order):
        self.layoutAboutToBeChanged.emit()

        self.xrefs = sorted(
            self.xrefs,
            key=lambda x: self._get_column_data(x, column),
            reverse=order == Qt.DescendingOrder,
        )
        self.layoutChanged.emit()

    def _get_column_text(self, xref, idx):
        if idx < len(self.HEADER):
            data = self._get_column_data(xref, idx)
            if isinstance(data, int):
                return hex(data)
            return data

    #
    # Abstract methods
    #

    def _get_column_data(self, xref, idx):
        raise NotImplementedError


class QXRefVariableModel(QXRefModel):
    HEADER = [("Direction", 70), ("Type", 50), ("Variable", 80), ("VarId", 80), ("PC", 160), ("Text", 300)]

    DIRECTION_COL = 0
    TYPE_COL = 1
    VARIABLE_COL = 2
    VARID_COL = 3
    PC_COL = 4
    TEXT_COL = 5

    def _get_column_data(self, ref, idx):
        """

        :param VariableAccess ref:
        :param idx:
        :return:
        """
        mapping = {
            self.DIRECTION_COL: self._direction,
            self.TYPE_COL: self._access_type,
            self.VARIABLE_COL: self._varname,
            self.VARID_COL: self._identstr,
            self.PC_COL: self._addrstr,
            self.TEXT_COL: self._text,
        }

        handler = mapping.get(idx)
        if handler is not None:
            return handler(ref)
        return None

    def _addrstr(self, _r):
        addr = _r.location.ins_addr
        node = self.instance.cfg.get_any_node(addr, anyaddr=True)
        if node is not None and node.function_address is not None:
            return f"{addr:#x} ({self.instance.kb.functions[node.function_address].name})"
        return hex(addr)

    def _direction(self, _r):
        """

        :param VariableAccess _r:
        :return:
        """
        if _r.location.ins_addr < self.addr:
            return "up"
        elif _r.location.ins_addr > self.addr:
            return "down"
        return ""

    @staticmethod
    def _access_type(_r):
        """

        :param VariableAccess _r:
        :return:
        """
        return _r.access_type

    @staticmethod
    def _varname(_r):
        """

        :param VariableAccess _r:
        :return:
        """
        return _r.variable.name

    @staticmethod
    def _identstr(_r):
        """

        :param VariableAccess _r:
        :return:
        """
        return _r.variable.ident

    def _text(self, _r):
        """

        :param VariableAccess _r:
        :return:
        """
        text = self.instance.get_instruction_text_at(_r.location.ins_addr)
        if text is None:
            text = "-=unavailable=-"
        return text


class QXRefAddressModel(QXRefModel):
    HEADER = [("Direction", 70), ("Type", 50), ("PC", 160), ("Text", 300)]

    DIRECTION_COL = 0
    TYPE_COL = 1
    PC_COL = 2
    TEXT_COL = 3

    def _get_column_data(self, ref, idx):
        """

        :param XRef ref:
        :param int idx:
        :return:
        """
        mapping = {
            self.DIRECTION_COL: self._direction,
            self.TYPE_COL: self._access_type,
            self.PC_COL: self._addrstr,
            self.TEXT_COL: self._text,
        }

        handler = mapping.get(idx)
        if handler is not None:
            return handler(ref)
        return None

    def _addrstr(self, _r):
        addr = _r.ins_addr
        node = self.instance.cfg.get_any_node(addr, anyaddr=True)
        if node is not None and node.function_address is not None:
            return f"{addr:#x} ({self.instance.kb.functions[node.function_address].name})"
        return hex(addr)

    def _direction(self, _r):
        """

        :param XRef _r:
        :return:
        """
        if _r.ins_addr < self.addr:
            return "up"
        elif _r.ins_addr > self.addr:
            return "down"
        return ""

    @staticmethod
    def _access_type(_r):
        """

        :param XRef _r:
        :return:
        """
        return _r.type_string

    def _text(self, _r):
        """

        :param XRef _r:
        :return:
        """
        text = self.instance.get_instruction_text_at(_r.ins_addr)
        if text is None:
            text = "-=unavailable=-"
        return text


class QXRefViewer(QTableView):
    def __init__(
        self,
        addr=None,
        variable_manager=None,
        variable=None,
        xrefs_manager=None,
        dst_addr=None,
        instance: "Instance" = None,
        xref_dialog=None,
        parent=None,
    ):
        super().__init__(parent)

        self.verticalHeader().setVisible(False)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setShowGrid(False)

        self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(24)

        self._instance = instance
        self._xref_dialog = xref_dialog
        self._addr = addr

        # Two modes
        # Mode A: (local-)variable-based
        self._variable_manager = variable_manager
        self._variable = variable
        # Mode B: Global address based
        self._xrefs_manager = xrefs_manager
        self._dst_addr = dst_addr

        # Determine which mode we are at
        if self._variable_manager is not None and self._variable is not None:
            self.mode = XRefMode.Variable
        elif self._xrefs_manager is not None and self._dst_addr is not None:
            self.mode = XRefMode.Address
        else:
            raise ValueError("Unsupported mode. Either variable or dst_addr should be specified.")

        self.items = []
        self._reload()

        if self.mode == XRefMode.Variable:
            self._model = QXRefVariableModel(self._addr, self._instance, self)
        elif self.mode == XRefMode.Address:
            self._model = QXRefAddressModel(self._addr, self._instance, self)
        else:
            raise ValueError("Unsupported mode. Either variable or dst_addr should be specified.")
        self.setModel(self._model)

        # set initial column widths
        for idx, (_, width) in enumerate(self._model.HEADER):
            self.setColumnWidth(idx, width)

        self.setSortingEnabled(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)

        hheader = self.horizontalHeader()
        hheader.setStretchLastSection(True)
        hheader.setSectionResizeMode(0, QHeaderView.ResizeToContents)

        self.doubleClicked.connect(self._on_item_doubleclicked)

    def _reload(self):
        if self.mode == XRefMode.Variable:
            self.items = self._variable_manager.get_variable_accesses(self._variable, same_name=True)
            self.items = sorted(self.items, key=lambda item: item.location.ins_addr)
        else:  # self.mode == XRefMode.Address
            self.items = list(self._xrefs_manager.get_xrefs_by_dst(self._dst_addr))

            # Create XRef objects for addresses inside the CFG
            self.items += list(self._xrefs_from_control_flow_transitions())
            self.items = sorted(self.items, key=lambda item: item.ins_addr)

    def _xrefs_from_control_flow_transitions(self):
        if self._instance is not None:
            cfg = self._instance.cfg
            node = cfg.get_any_node(self._dst_addr)
            if node is not None:
                # its predecessors
                predecessors = cfg.get_predecessors(node)
                arch = self._instance.project.arch
                for pred in predecessors:
                    if pred.instruction_addrs:
                        if arch.branch_delay_slot and len(pred.instruction_addrs) > 1:
                            ins_addr = pred.instruction_addrs[-2]
                        else:
                            ins_addr = pred.instruction_addrs[-1]
                    else:
                        ins_addr = pred.addr
                    yield XRef(
                        ins_addr=ins_addr,
                        block_addr=pred.addr,
                        stmt_idx=None,
                        dst=self._dst_addr,
                        xref_type=XRefType.Offset,
                    )

    #
    # Signal handlers
    #

    def _on_item_doubleclicked(self, model_index):
        row = model_index.row()
        xref_dialog = self._xref_dialog
        if xref_dialog is None:
            return

        item = self.items[row]
        if isinstance(item, VariableAccess):
            xref_dialog.jump_to(item.location.ins_addr)
        elif isinstance(item, XRef):
            xref_dialog.jump_to(item.ins_addr)



import itertools
import json
import random
import re
import string
from collections import defaultdict
from typing import Set, Tuple

import requests
from PySide6.QtGui import Qt
from PySide6.QtWidgets import QMessageBox
from sortedcontainers import SortedDict

from angrmanagement.config import Conf
from angrmanagement.plugins.base_plugin import BasePlugin


class VaRec(BasePlugin):
    """
    The plugin for supporting the VaRec plugin (private for now until it is released to the public).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.transitions: Set[Tuple[int, int]] = set()
        self.covered_blocks = SortedDict()

        self.sink_color = Qt.yellow

    MENU_BUTTONS = [
        "&Infer variable names",
    ]
    INFER_VARIABLE_NAMES = 0

    def handle_click_menu(self, idx):
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        if self.workspace.main_instance.project.am_none:
            return

        mapping = {
            VaRec.INFER_VARIABLE_NAMES: self.infer_variable_names,
        }

        mapping.get(idx)()

    @staticmethod
    def _restore_stage(view):
        # shrug
        for v in view.codegen._variable_kb.variables[view.function.addr]._unified_variables:
            m = re.match(r"@@(\S+)@@(\S+)@@", v.name)
            if m is not None:
                var_name = m.group(1)
                v.name = var_name
        # refresh the view
        view.codegen.regenerate_text()
        view.codegen.am_event()

    @staticmethod
    def randstr(n=8):
        return "".join(random.choice(string.ascii_lowercase) for _ in range(n))

    def infer_variable_names(self):
        view = self.workspace._get_or_create_pseudocode_view()
        if view.codegen.am_none:
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Cannot predict variable names. No pseudocode exists in the pseudocode view.",
                QMessageBox.Ok,
            )
            return
        if view.codegen._variable_kb is None:
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Cannot predict variable names. The pseudocode view does not have associated variables KB.",
                QMessageBox.Ok,
            )
            return

        proxies = {"http": Conf.http_proxy, "https": Conf.https_proxy} if Conf.http_proxy or Conf.https_proxy else None

        for v in view.codegen._variable_kb.variables[view.function.addr]._unified_variables:
            if not v.renamed:
                v.name = f"@@{v.name}@@{VaRec.randstr()}@@"

        view.codegen.regenerate_text()
        d = {
            "code": [
                {
                    "raw_codes": [
                        view.codegen.text,
                    ]
                }
            ]
        }
        r = requests.post(f"{Conf.varec_endpoint}", data=json.dumps(d), proxies=proxies)
        try:
            result = json.loads(r.text)
        except json.JSONDecodeError:
            self._restore_stage(view)

            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Failed to predict names for all variables involved.",
                QMessageBox.Ok,
            )

            return

        varname_blacklist = {"UNK", "null", "true", "false", "return", "do", "while"}
        varname_to_predicted = defaultdict(list)

        # handle failure cases
        if "code" not in result or not result["code"]:
            self._restore_stage(view)
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Unexpected output returned from the backend. 'code' is not found or empty.",
                QMessageBox.Ok,
            )
            return
        if "predictions" not in result["code"][0] or not result["code"][0]["predictions"]:
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Unexpected output returned from the backend. 'predictions' is not found or empty.",
                QMessageBox.Ok,
            )
            self._restore_stage(view)
            return
        if len(result["code"][0]["predictions"]) == 1 and isinstance(result["code"][0]["predictions"][0], str):
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                f"Prediction failed. Error: {result['code'][0]['predictions'][0]}",
                QMessageBox.Ok,
            )
            self._restore_stage(view)
            return

        for idx, m in enumerate(re.finditer(r"@@(\S+)@@(\S+)@@", view.codegen.text)):
            var_name = m.group(1)
            prediction = result["code"][0]["predictions"][0][idx]
            topk = prediction["top-k"]
            # remove variable names that we don't like
            filtered_topk = [item for item in topk if item["pred_name"] not in varname_blacklist]
            if filtered_topk:
                varname_to_predicted[var_name].extend(filtered_topk)

        ctrs = defaultdict(itertools.count)

        # rename them all
        used_names = set()
        for v in view.codegen._variable_kb.variables[view.function.addr]._unified_variables:
            m = re.match(r"@@(\S+)@@\S+@@", v.name)
            if m is not None:
                var_name = m.group(1)
                predicted = varname_to_predicted[var_name]
                predicted = sorted(predicted, key=lambda x: x["confidence"], reverse=True)
                v.candidate_names = {pred["pred_name"] for pred in predicted}
                for pred in predicted:
                    if pred["pred_name"] not in used_names:
                        v.name = pred["pred_name"]
                        used_names.add(v.name)
                        break
                else:
                    if predicted:
                        v.name = predicted[0]["pred_name"] + "_" + str(next(ctrs[predicted[0]["pred_name"]]))
                    else:
                        v.name = var_name  # restore the original name
        view.codegen.am_event()
